// Module.h : main header file for TModule and related classes.
//

#if !defined(AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
#define AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
#error include 'stdafx.h' before including this file for PCH
#endif

#include <afxmt.h>
#include <afxtempl.h>
#include "sysdef.h"
#include "Utils/Semaphore.h"
#include "Utils/Time.h"
#include "Utils/Users.h"
#include "Xpert/Win32Fix.h"

#undef DECL
#ifdef ENGINE_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif


///////////////////////////////////////////////////////////////////////////////
// Module base class and its supporting classes
//

/*
Diagram which attempts to show the various screen attributes used for displaying the graphical
sensor hookup screen:

--------------------------------------------------------------               ---
(tm)                                                                             |
--------------------------------------------------------------                |
(it)                                                        |
+-----------+                  +-----------+ (iom)      ---     |
(lm)          ---|           |---  (rm)      ---|           |---          |      |
(mi) --|           |-----+         ---|           |--- (mo)     |      |
+-------|           |---  |    (ios)---|(il)   (ir)|---        (ih)   (ch)
|    ---|           |---  |         ---|           |---          |      |
|    ---|           |---  |      +-----|           |---          |      |
|       +-----------+     |      |     +-----------+ (iom)      ---     |
(ib)          +------+     <-----iw---->                    |
|
--------------------------------------------------------------                |
(bm)                                                                             |
--------------------------------------------------------------               ---
<---------------(cw)-------------------->
*/


/*

Module Property Design:

1) Properties are stored in a variant data structure so we can support multiple data types,
every data type can be converted to and from strings so we can write out a setup or edit.

2) Each derivation of TModule can define it's own list of Properties and Variants.

TPropertyList MyList[] = { {"Slope", &Slope}, {"Offset", &Offset} };
void MyModule::SetupProperties()
{
MyAncestor::SetupProperties();
AddProperties(MyList);
};

3) AddProperties would add each property to a MapStringToPtr structure to allow for
fast lookup by name.

4) Finally we need some functions to enumerate all the available properties to
support streaming them out to disk.

Note: there may be some overlap between properties and real time values which can
be eliminated.

*/


//////////////////////////////////////////////////////////////////////////
// The following constants are used in Module base class implementation

//in order to have more than 5 outputs for engine execution purposes
//we define a seperate number of max outputs, for sdisplay purposes
//there is still a number limited by the size of icons
namespace ENG
{
   const MaxOutputs = 20;
}

namespace UI
{
   const MaxInputs                = 5;  // (mi)
   const MaxOutputs               = 5;  // (mo)
   const ZoomLevels               = 2;
};

namespace UIPrivate
{
   const LabelHeight[UI::ZoomLevels]  = {10,10}; // Text label below the icon
   const LeftMargin[UI::ZoomLevels]   = {5, 10};  // (lm
   const RightMargin[UI::ZoomLevels]  = {5, 10};  // (mo)
   const TopMargin[UI::ZoomLevels]    = {5, 10};  // (mtr)
   const BottomMargin[UI::ZoomLevels] = {5 + LabelHeight[0], 5 + LabelHeight[1]};  // (mbr)
   const IconHeight[UI::ZoomLevels]   = {32, 84}; // (ih)
   const IconWidth[UI::ZoomLevels]    = {32, 84}; // (iw)
   const IOSpacing[UI::ZoomLevels]    = {IconHeight[0]/UI::MaxInputs, IconHeight[1]/UI::MaxInputs};                    // (ios)
   const IOMargin[UI::ZoomLevels]     = {IconHeight[0]%UI::MaxInputs/2, IconHeight[1]%UI::MaxInputs/2};    // (iom)
   const IconTop[UI::ZoomLevels]      = {TopMargin[0],TopMargin[1]};             // (it)
   const IconBottom[UI::ZoomLevels]   = {IconTop[0] + IconHeight[0] - 1, IconTop[1] + IconHeight[1] - 1};        // (ib)
   const IconLeft[UI::ZoomLevels]     = {LeftMargin[0], LeftMargin[1]};          // (il)
   const IconRight[UI::ZoomLevels]    = {IconLeft[0] + IconWidth[0] - 1,IconLeft[1] + IconWidth[1] - 1};                // (ir)
   const CellWidth[UI::ZoomLevels]    = {IconWidth[0] + LeftMargin[0] + RightMargin[0], IconWidth[1] + LeftMargin[1] + RightMargin[1]};    // (cw)
   const CellHeight[UI::ZoomLevels]   = {IconHeight[0] + TopMargin[0] + BottomMargin[0], IconHeight[1] + TopMargin[1] + BottomMargin[1]};   // (ch)
};


class DECL TValue
{
public:
   TValue()                               {DataType=0; RightDigits=-15;};
   ~TValue()                              {};
   TValue(TCHAR* Str)                     {*this = Str; RightDigits=-15;};
   TValue(LPCTSTR Str)                    {*this = Str; RightDigits=-15;};
   TValue(const CString Str)              {*this = Str; RightDigits=-15;};
   TValue(int Num)                        {*this = Num; RightDigits=-15;};
   TValue(bool Num)                       {*this = Num; RightDigits=-15;};
   TValue(double Num)                     {*this = Num; RightDigits=-15;};
   TValue& operator=(const TValue& o)     {DataType=o.DataType; Integer=o.Integer; Double=o.Double; String=o.String; RightDigits=o.RightDigits; return *this; };
   TValue& operator=(TCHAR* Str)          {DataType=STRING; String=Str; return *this;         };
   TValue& operator=(LPCTSTR Str)         {DataType=STRING; String=Str; return *this;         };
   TValue& operator=(const CString Str)   {DataType=STRING;  String=Str; return *this;        };
   TValue& operator=(int Num)             {DataType=INTEGER; Integer=Num; return *this;       };
   TValue& operator=(bool Num)            {DataType=INTEGER; Integer=(int) Num; return *this; };
   TValue& operator=(double Num)          {DataType=DOUBLE; Double=Num; return *this;         };
//
   TValue GetArray(int Index);
   void SetArray(int Index, const TValue& Value);
   TValue operator[](int Index)           {return GetArray(Index);                 };
   operator int()                         {MakeInt(); return Integer;              };
   operator unsigned int()                {MakeInt(); return Integer;              };
   operator bool()                        {MakeInt(); return Integer != 0;         };
   operator CString()                     {MakeStr(); return String;               };
   operator const CString&()              {MakeStr(); return String;               };
   operator double()                      {MakeDbl(); return Double;               };
//
   TValue& operator+=(TValue& v);
   TValue& operator+=(double d)           {MakeDbl(); Double += d; return *this;   };
   TValue& operator+=(int n)              {MakeInt(); Integer += n; return *this;  };
   TValue& operator+=(CString s)          {MakeStr(); String += s; return *this;   };
//
   bool operator==(TValue& v);
   bool operator==(double d)               {MakeDbl(); return (Double == d);       };
   bool operator==(int n)                  {MakeInt(); return (Integer == n);      };
   bool operator==(CString s)              {MakeStr(); return (String == s);       };
//
   bool operator!=(TValue& v);
   bool operator!=(double d)               {MakeDbl(); return (Double != d);       };
   bool operator!=(int n)                  {MakeInt(); return (Integer != n);      };
   bool operator!=(CString s)              {MakeStr(); return (String != s);       };
//
   bool operator<(TValue& v);
   bool operator<(double d)               {MakeDbl(); return (Double < d);         };
   bool operator<(int n)                  {MakeInt(); return (Integer < n);        };
   bool operator<(CString s)              {MakeStr(); return (String < s);         };
//
   bool operator>(TValue& v);
   bool operator>(double d)               {MakeDbl(); return (Double > d);         };
   bool operator>(int n)                  {MakeInt(); return (Integer > n);        };
   bool operator>(CString s)              {MakeStr(); return (String > s);         };
//
   bool operator<=(TValue& v);
   bool operator<=(double d)              {MakeDbl(); return (Double <= d);        };
   bool operator<=(int n)                 {MakeInt(); return (Integer <= n);       };
   bool operator<=(CString s)             {MakeStr(); return (String <= s);        };
//
   bool operator>=(TValue& v);
   bool operator>=(double d)              {MakeDbl(); return (Double >= d);        };
   bool operator>=(int n)                 {MakeInt(); return (Integer >= n);       };
   bool operator>=(CString s)             {MakeStr(); return (String >= s);        };
//
   // I didn't implement an operator LPCTSTR because of conversion conflicts w/CString
   LPCTSTR AsLPCTSTR()                    {MakeStr(); return (LPCTSTR) String;     };
   int& AsInteger()                       {MakeInt(); return Integer;              };
   CString& AsCString()                   {MakeStr(); return String;               };
   double& AsDouble()                     {MakeDbl(); return Double;               };
   bool IsInteger() const                 {return (DataType & IS_INT) != 0;        };
   bool IsDouble() const                  {return (DataType & IS_DBL) != 0;        };
   bool IsString() const                  {return (DataType & IS_STR) != 0;        };
   bool IsNull() const                    {return DataType == 0;                   };
   void SetRightDigits(int Digits)        {MakeDbl(); DataType=DOUBLE; RightDigits = Digits;};
   int GetRightDigits() const             {return RightDigits;                     };
   void SetType(int Type)                 {DataType = Type;                        };
   int GetType() const                    {return DataType;                        };
protected:
   void MakeInt()                         {if ((DataType & CVT_INT)==0) {if (DataType & DOUBLE) Integer=(int) Double; else if (DataType & STRING) Integer=_ttoi(String); else Integer=0; DataType|=CVT_INT;}; };
   void MakeStr()                         {if ((DataType & CVT_STR)==0) {if (DataType & DOUBLE) {if (RightDigits < 0) String.Format(_T("%.*g"), -RightDigits, Double); else String.Format(_T("%1.*f"), RightDigits, Double);} else if (DataType & INTEGER) String.Format(_T("%d"), Integer); else String=_T(""); DataType|=CVT_STR;}}
   void MakeDbl()                         {if ((DataType & CVT_DBL)==0) {if (DataType & INTEGER) Double=Integer; else if (DataType & STRING) Double=_tcstod(String, NULL); else Double=0.0; DataType|=CVT_DBL;}; };
protected:
   enum TDataType {IS_INT=1, IS_DBL=2, IS_STR=4, CVT_INT=8, CVT_DBL=16, CVT_STR=32, INTEGER=IS_INT|CVT_INT, DOUBLE=IS_DBL|CVT_DBL, STRING=IS_STR|CVT_STR};
   int DataType;
   int Integer;
   double Double;
   CString String;
   int RightDigits;
};


class DECL TProperty : public TValue
{
public:
   TProperty()                              {Attributes=0; };
   TProperty& operator=(const TProperty& o) {Attributes=o.Attributes; TValue::operator=(o); return *this;};
   TProperty& operator=(const TValue& o)    {TValue::operator=(o);  return *this;};
   TProperty& operator=(TCHAR* Str)         {TValue::operator=(Str); return *this;};
   TProperty& operator=(LPCTSTR Str)        {TValue::operator=(Str); return *this;};
   TProperty& operator=(const CString Str)  {TValue::operator=(Str); return *this;};
   TProperty& operator=(int Num)            {TValue::operator=(Num); return *this;};
   TProperty& operator=(bool Num)           {TValue::operator=(Num); return *this;};
   TProperty& operator=(double Num)         {TValue::operator=(Num); return *this;};
public:
   enum TAttributes {NOSAVE=1, READONLY=2, SYSTEM=4, DYNAMIC=8};
   int Attributes;
};


/////////////////////////////////////////////////////////////////////////////
// TMapStringToRef is an (IMO) improved version of TMapStringToPtr that works with references instead of ptrs,
// and is type safe by using templates - simplified from CMapStringToPtr.
template<class TYPE>
class TMapStringToRef : public CMapStringToPtr
{
public:
   BOOL Lookup(LPCTSTR key, TYPE*& V) const                    { return CMapStringToPtr::Lookup(key, (void *&) V);    };
   TYPE& operator[](LPCTSTR key)                               { return * (TYPE*&) CMapStringToPtr::operator[](key);  };
   void SetAt(LPCTSTR key, TYPE& V)                            { CMapStringToPtr::SetAt(key, &V);              };
   void GetNextAssoc(POSITION& P, CString& K, TYPE*& V) const  { CMapStringToPtr::GetNextAssoc(P, K, (void *&) V);    };
};

/////////////////////////////////////////////////////////////////////////////
// TMapStringToRefUpperCase is a version of TMapStringToRef that uppercases all strings
template<class TYPE>
class TMapStringToRefUpperCase : public CMapStringToPtr
{
public:
   BOOL Lookup(CString K, TYPE*& V) const                      { K.MakeUpper(); return CMapStringToPtr::Lookup(K, (void *&) V);    };
   TYPE& operator[](CString K)                                 { K.MakeUpper(); return * (TYPE*&) CMapStringToPtr::operator[](K);  };
   void SetAt(CString K, TYPE& V)                              { K.MakeUpper(); CMapStringToPtr::SetAt(K, &V);                     };
   void GetNextAssoc(POSITION& P, CString& K, TYPE*& V) const  { CMapStringToPtr::GetNextAssoc(P, K, (void *&) V);                 };
   BOOL RemoveKey(CString K)                                   { K.MakeUpper(); return CMapStringToPtr::RemoveKey(K);              };
};

class TPropertyList : public TMapStringToRef<TProperty>
{
public:
   void Add(const CString Str, TProperty& Property)  { SetAt(Str, Property); };
   void Add(const CString Str, TProperty& Property, TProperty::TAttributes Attr)  { Property.Attributes=Attr; Add(Str, Property); };
   bool Remove(const CString Str)                    { return RemoveKey(Str) != 0; };
   bool Get(const CString Str, TProperty*& Property) { return Lookup(Str, Property) != 0; };
   bool GetNext(POSITION& P, CString& Str, TProperty*& Property)     { if (P) { GetNextAssoc(P, Str, Property); return true; } else return false; };
private:
   TPropertyList* List;
};



/////////////////////////////////////////////////////////////////////////////
// TObArray is an (IMO) improved version of CObArray that works with references instead of ptrs,
// and is type safe by using templates - simplified from CTypedPtrArray.
template<class TYPE>
class TObArray : public CObArray
{
public:
   TYPE& GetAt(int nIndex) const               { return (TYPE&) *CObArray::GetAt(nIndex);     };
   TYPE& ElementAt(int nIndex)                 { return (TYPE&) CObArray::ElementAt(nIndex);  };
   void SetAt(int nIndex, TYPE& newElement)    { CObArray::SetAt(nIndex, &newElement);        };
   void SetAtGrow(int nIndex, TYPE& newElement){ CObArray::SetAtGrow(nIndex, &newElement);    }
   int Add(TYPE& newElement)                   { return CObArray::Add(&newElement);           };
   int Append(const TObArray<TYPE>& src)       { return CObArray::Append(src);                };
   void Copy(const TObArray<TYPE>& src)        { CObArray::Copy(src);                         };
   void InsertAt(int nIndex, TYPE& newElement, int nCount = 1) { CObArray::InsertAt(nIndex, &newElement, nCount); };
   void InsertAt(int nStartIndex, TObArray<TYPE>& NewArray)    { CObArray::InsertAt(nStartIndex, &NewArray);      };
   TYPE& operator[](int nIndex) const          { return (TYPE&) *CObArray::operator[](nIndex);        };
};


///////////////////////////////////////////////////////////////////////////////
// CModPropPage - wrapper class used by modules that create property pages.
//
class DECL CModPropPage : public CPropertyPage
{
public:
   // Constructors
   CModPropPage(UINT nIDTemplate, UINT nIDCaption = 0)
      : CPropertyPage(nIDTemplate, nIDCaption) { }
   CModPropPage(LPCTSTR lpszTemplateName, UINT nIDCaption = 0)
      : CPropertyPage(lpszTemplateName, nIDCaption) { }

   // Attributes
   TUserGroup m_ug;           // The group this module is in
   TUserGroup LoginGroup;   // The group user logged in as
};


///////////////////////////////////////////////////////////////////////////////
// TNode represents a connection point between modules.

//forward declare
class TModule;
typedef TModule * PModule;


//a node basically manages the connection point and a ptr to the module connected
class DECL TNode : public CObject
{
public:
   TNode()                                { Clear();};
   TNode(TModule& AModule, int APoint)    { Module=&AModule; point=APoint;      };
   TNode(const TNode& n)                  { Module=n.Module; point=n.point;     };
   virtual ~TNode()                       {};
   bool operator==(const TNode& n) const  { return ((n.Module==Module) && (n.point==point)); };
   bool operator!=(const TNode& n) const  { return ((n.Module!=Module) || (n.point!=point)); };
   TNode& operator=(const TNode& n)       { Module=n.Module; point=n.point; return *this;    };
public:
   void Clear()                           { Module=NULL; point=-1;              };
   bool Empty()                           { return Module==NULL;                };
   bool Inuse()                           { return Module!=NULL;                };
   TNode& Source();            // returns node which is the source of this node
   bool IsInput()                         { return (point >= 0) && (point < UI::MaxInputs);  };
   bool IsOutput()                        { return (point >= 0) && (point >= UI::MaxInputs);  };
   void SetInput(int Pt)                  { point = Pt;                         };
   void SetOutput(int Pt)                 { point = Pt + UI::MaxInputs;         };

   //Modified for "more" output capability, Do we want to keep this?
   //there are 2 types of points now, one for display of outputs in icon and
   //another for execution of engine.
   int Point();//  { return (point < UI::MaxInputs) ? point : point-UI::MaxInputs; };
   int EngPoint();

public:
   PModule Module;
private:
   int point;
};


/////////////////////////////////////////////////////////////////////////////
// CSensorData, used to pass data among modules

#define realtype double

class DECL CSensorData : public CObject
{
public:
   virtual ~CSensorData();
   CSensorData(const CSensorData& data);
   CSensorData();
   BOOL operator == (const CSensorData& data) const;
   CSensorData& operator=(const CSensorData& data);
   void SetID(CString& string){SensorID = string;};
   CString& GetID(){return SensorID;};
   CString FormatData();
   CString FormatData(int iRightDigits);
   CString GetQualStr();

   //what are valid quality types??
   enum QualityType {GOOD, BAD, UNDEFINED};
// 9000 SSP Alarm Status
   enum TAlarmStatus  // (* Controls how alarms are handled, and current status *)
   {
// Alarm Status Word
      HiLimitA        = 1<<0,
      LowLimitA       = 1<<1,
      IntHiLimitA     = 1<<2,
      IntLowLimitA    = 1<<3,
      ControlTimeoutA = 1<<4,
      ControlFailA    = 1<<5,
      ControlSoftA    = 1<<6, 
      ControlHardA    = 1<<7,
      DigitalA        = 1<<8,
      InhibitA        = 1<<9,
// Change Status Word
      HiLimitC        = 1<<16,
      LowLimitC       = 1<<17,
      IntHiLimitC     = 1<<18,
      IntLowLimitC    = 1<<19,
      ControlTimeoutC = 1<<20,
      ControlFailC    = 1<<21,
      ControlSoftC    = 1<<22, 
      ControlHardC    = 1<<23,
      DigitalC        = 1<<24,
      InhibitC        = 1<<25
   };

   // Overrides
   // ClassWizard generated virtual function overrides
   //{{AFX_VIRTUAL(CSensorData)
   //}}AFX_VIRTUAL

   // Implementation
public:
   CTime           TimeScheduled; //scheduled timestamp
   CTime           TimeActual;    //actual timestamp
   CString         SensorID;      //id of sensor supplying this data
   TValue          Data;          //data itself
   CString         Units;         //how to track units??
   QualityType     Quality;       //what is quality of this data
   DWORD           AlarmStatus;   //Alarm status of this value (bit fields)
   UINT32          DeviceTime;
};


enum CIODeviceType {UNKNOWN_DEVICE, ANALOG, DIGITAL, DISPLAY, SDI12, RS232, RS485};

//////////////////////////////////////////////////////////////////////////////////
// TModule, the module base class that all other modules should be derived from
class CXMLSetup;        //fwd declare
class CXMLToken;

class DECL TModule : public CObject
{
   friend class TPropertyEnum;
public:
   TModule(const CString& ModuleName, HINSTANCE hModInst);
   virtual ~TModule();
   ////////////////////////////////////////////////////////////////////////
   //The following functions are used for managing the GUI parts of modules

   int FindInput(TModule& Module);              // Returns Inputs[] index if Module is an input else -1.
   int FindInput(TNode Node);                   // Returns the Inputs[] index of an input connected to Node else -1
   int WiredInputCount();                       // Returns the number of inputs which are wired up.
   void RemoveInput(TModule& Module);           // Removes any inputs matching Module
   BOOL IsPtInUse(int nPt, TModule* pModule);
   int GetConnectionPoint(TModule* pSrcMod);
   int FindModuleIndex(TModule* pMod, TObArray<TModule>& ModList);

   // Virtual function that may be overridden if desired
   virtual int GetDefaultInput(bool GetUnused=false, int Index=0);  // Pick a default input for a just created module
   virtual int GetDefaultOutput(int Index=0);       // Pick a default output for a just created module
   virtual bool ShowProperties(CWnd* pParent) {return false;};   // Show edit properties dialog. Return true on change.
   virtual int Rename(CWnd* pParent);               // Get a new function name from user.
   virtual bool GetIOAddrAsString(CString& string); //get a string describing I/O location, return false if unassigned
   virtual bool GetSensorInfo(CArray<CString, CString&>& info); //get strings to display in View Sensors, return false if no data should be displayed
   virtual CString OutputNamePopup(int Point){return OutputName(Point);};
   virtual CBitmap& GetBitmap(CBitmap& bm);     // Loads & returns a 32x32 bitmap to represent the module
   virtual void OnWireChange(CWnd* pParent) {}; // Called when change to module wiring occurs.
   virtual void AssignDefaultChannel() {};      // Called to setup default wiring for a new module.
   virtual void ClearChannelsInuse()  {};       // Should clear any I/O channel assignments
   virtual void SetChannelsInuse()    {};       // Should make any necessary I/O channel assignments
   virtual CIODeviceType GetDeviceType()          // Returns the type of device assuming it's an I/O module
      {return UNKNOWN_DEVICE;};
   virtual TProperty* GetModuleProperty()       // Returns the property that controls the I/O module number, Port, or Address
      {return NULL;};
   virtual TProperty* GetChannelProperty()      // Returns the property that controls the channel number
      {return NULL;};
   virtual TProperty* GetSlopeProperty(int Point);  // Returns the slope property associated with an output point
   virtual TProperty* GetOffsetProperty(int Point); // Returns the offset property associated with an output point

   //pure virtuals that must be defined for each module
   virtual CString InputName(int Point) = 0;    // Returns the name of an Input
   virtual CString OutputName(int Point) = 0;   // Returns the name of an Output
   virtual bool InputActive(int Point) = 0;     // Returns true if a module uses an input point
   virtual bool OutputActive(int Point) = 0;    // Returns true if a module uses an output point
   virtual TModule* CreateAnother() = 0;        // Calls default constructor to create another module of the same type

   //////////////////////////////////////////////////////////////////////////////
   //These functions are for support of reading/writing modules into setup files
private:
   TPropertyList PropertyList;

public:
   void AddProperty(const CString Str, TProperty& Property)  { PropertyList.Add(Str, Property);         };
   void AddProperty(const CString Str, TProperty& Property, TProperty::TAttributes Attr)  { PropertyList.Add(Str, Property, Attr); };
   bool RemoveProperty(const CString Str)                    { return PropertyList.Remove(Str); };
   bool GetProperty(const CString Str, TProperty*& Property) { return PropertyList.Get(Str, Property); };

   // Override to prohibit dynamic property creation.
   virtual bool SupportsDynamicProperties(void) { return true; };

public:
   void WriteXMLSetup(CXMLSetup& XMLSetup, TObArray<TModule>& ModuleList, BOOL fDefaults);
   BOOL ReadXMLSetup(CXMLSetup& XMLSetup, CXMLToken& XMLToken, TObArray<TModule>& ModuleList, int BaseIndex);

   //virtual functions that may be overridden
   virtual void AfterReadXMLSetup(int BaseIndex);
   virtual void AfterReadSetup(int BaseIndex);
   virtual void BeforeWriteSetup();
   virtual bool IsActive() { return false; };            // True when block initiates data flow (e.g., Measure).
   virtual CString GetScheduleInfo() { return _T(""); }; // Return textual description of schedule (when module is active).

public: // GUI related publics that need to be set
   virtual int InputCount() = 0;                // Number of active inputs & outputs
   virtual int OutputCount() = 0;               // Number of active outputs
   CString Name;                                // Name of the Module. (i.e. "Measure")

public: // GUI related publics that are calculated.
   int X, Y;                                    // Position of ICON on the screen
   bool bSelected;                              // GUI icon is selected for display.

public:
   //This node array tracks the input connections to this module
   TNode Inputs[UI::MaxInputs];

   //This index array tracks inputs but without the Module ptr.
   //Needed for saving the adjacent modules in the setup file
   int Indices[UI::MaxInputs];

   /////////////////////////////////////////////////////////////////////////////////
   //additions for support of run-time engine.

   //list of i2c event types
   enum I2CEvent {ALARM = 0x01, OVERSPEED = 0x02};

   //virtual functions that may be overridden
   virtual void Initialize() {};                             // to prepare for running engine
   bool Push(TTime tScheduled);                              // Execute and push the result to the outputs
   bool Pull(TTime tScheduled, bool bLiveReading);           // Execute inputs and compute new result. Set bLiveReading to true to override InhibitPull logic.
   void Get(int Point, CSensorData& Data);                   // Get Last Measured Value for a Point
   void Set(int Point, const CSensorData& Data);             // Set the Last Measured Value for a Point
   bool GetInputData(int Point, CSensorData& Data);          // Get Last Measure Value from an Input Module
   bool SetInputData(int Point, CSensorData& Data);          // Set Last Measure Value from an Input Module
   void KickBack(CSensorData& Data);                         // Kick a modified value back to anyone interested in it

   virtual int GetModuleType() {return TModule::INPUT;};     // what type of module is this (input, telemetry etc)
   virtual void Stop() {};                                   // for handling any cleanup on stop
   virtual void EventExec(BYTE EventType, UINT32 Time) {};
   virtual bool Calibrate(void);                             // Override with actual calibration routine. It should return true to force setup save.
   virtual bool I2CCalibrate(void) {return false;};          // Override with actual calibration routine. It should return true to force setup save.
   virtual bool SupportsI2CCalibrate(void);                  // Override if there is a cal routine
   virtual void ExecuteScheduledAction(TTime tScheduled, bool bEvent) { Pull(tScheduled, bEvent); };   // Hook to provide action for schedule object.
   virtual bool InhibitPull(TTime tScheduled);              // Hook to define when Pull should be inhibited. Ignored if a force is in effect.
   //list of module types
   enum ModuleType {INPUT, OUTPUT, PROCESSING, CALCULATION, LOGGING, TELEMETRY, OTHER};
   TObArray<TNode> Outputs;

protected: // Internal implementations of Public functions:
   virtual void Execute(TTime tScheduled) {};               // Compute current values
   virtual void Execute(TTime tScheduled, bool bLiveReading) { Execute(tScheduled); }; // Compute current values but make bLiveReading flag available
   virtual void PushData(TTime tScheduled);                                            // An event is pushing data
   virtual void PullData(TTime tScheduled, bool bLiveReading);                         // The module is being polled for data
   virtual void SetData(int Point, const CSensorData& Data)  { LastData = Data; };     // Set the Last Measured Value for a Point
   virtual void GetData(int Point, CSensorData& Data)        { Data = LastData; };     // Get Last Measured Value for a Point
   virtual void OnKickBack(CSensorData& Data) {};            // Some modules modify data then kick it back so previous modules can store the modified value if they wish
   virtual DWORD GetLiveReadingTimeout()                     { return 5000;     };     // How long to wait before aborting a livereading (0 means LiveReadings shouldn't be taken)

protected:
   CSensorData LastData;   // Default LastData

public: // Status information
   TTime tLastExecute;     // Absolute time of last execute.
   TTime tLastScheduled;   // Time of last scheduled execution (set by CSchedule)
   TTime tNextScheduled;   // Time of next scheduled execution (set by CSchedule)

protected: // Helper functions
// bool GetInputData(int Point, CSensorData& Data);               // Get Last Measure Value from an Input Module
   bool PullInputs(TTime tScheduled, bool bLiveReading);          // Pull Data from all Inputs
   bool PushOutputs(TTime tScheduled);                            // Push Data to all outputs
   bool PushOutputs(TTime tScheduled, int Output);                // Push Data to a specific output
   void LockData();                                               // Prevent access to Data
   void UnlockData();                                             // Allow access to Data

protected:
   TSemaphore PullLock;                            // for thread-safe access of Pull's
   TSemaphore PushLock;                            // for thread-safe access of Push'sll's
   CCriticalSection DataLock;                      // For GetData(), SetData() calls
   HINSTANCE ModuleInstance;

public: // PROPERTIES
   TProperty Function;                       // Function might be something like "Measure Stage", it's intended as a GUI label
   TProperty Input[UI::MaxInputs];
   TProperty Index;
   TProperty IsGUI;                           // True for modules editable with the setup GUI
};


// Help class for enumerating property lists.
class TPropertyEnum
{
public:
   TPropertyEnum(TModule& Module)                   { List = &Module.PropertyList; P = List->GetStartPosition(); };
   bool Get(CString& Str, TProperty*& Property)     { return List->GetNext(P, Str, Property);                    };
private:
   TPropertyList* List;
   POSITION P;
};


///////////////////////////////////////////////////////////////////////////
//Common Sutron DLL interface class
//
//This is a class which creates and describes TModule's and their PropertyPages.
class DECL TModuleFactory : public CObject
{
public:
   TModuleFactory()                               { ModuleCount = 0; ModuleList = NULL; PageCount = 0; PageList = NULL; };
   virtual ~TModuleFactory()                      {};
public: // Every DLL should over-ride the following
   virtual TModule* CreateModule(int Index) = 0;
   virtual CModPropPage* CreatePage(int Index) = 0;
public: // Only need to be over-ridden if the DLL's constructor doesn't initalize ModuleCount, ModuleList, PageCount, PageList
   virtual int GetModuleCount()                   { return ModuleCount;                    };
   virtual CString GetModuleName(int Index)       { return (CString) ModuleList[Index];    };
   virtual int GetPageCount()                     { return PageCount;                      };
   virtual CString GetPageName(int Index)         { return (CString) PageList[Index];      };
public:
   int FindModule(const CString& Name);
   int FindPage(const CString& Name);
   TModule* CreateModule(const CString& Name)     { return CreateModule(FindModule(Name)); };
   CModPropPage* CreatePage(const CString& Name) { return CreatePage(FindPage(Name));     };
protected:
   int ModuleCount;
   int PageCount;
   TCHAR** ModuleList;
   TCHAR** PageList;
};


//This is a base class for modules that need to add an entry to
//the NextGen Control Panel tree
class DECL CControlPanelEntry : public CObject
{
public:
   CControlPanelEntry() {m_pTree = NULL; m_hRootItem = NULL; m_bRecStopped = false; m_bSaveNeeded = false;};
   virtual void InsertItems(CTreeCtrl * pTree) {};
   virtual bool EditItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual bool DeleteItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual bool NewItem(CWnd* pParent, HTREEITEM hItem, CTreeCtrl* pTree) {return false;};
   virtual void GetButtonLabels(HTREEITEM hItem, CString& NewButton, CString& EditButton, CString& DeleteButton) {};
   virtual void OnKillActive(CWnd* pParent) {};
   virtual bool PromptStopEngine();
   CTreeCtrl*  m_pTree;
   HTREEITEM m_hRootItem;
   bool m_bRecStopped;
   bool m_bSaveNeeded;
};


class DECL TDialog : public CDialog
{
public:
   TDialog() : CDialog() {};
   TDialog(LPCTSTR lpszTemplateName, CWnd* pParentWnd = NULL) : CDialog(lpszTemplateName, pParentWnd) {};
   TDialog(UINT nIDTemplate, CWnd* pParentWnd = NULL) : CDialog(nIDTemplate, pParentWnd) {};
public:
   virtual LRESULT WindowProc(UINT message, WPARAM wParam, LPARAM lParam );
private:
   bool TimerKludgeFix;
};


////////////////////////////////////////////////////////////////////////////
// Global utility functions, called by indiviual modules

// SetTimeDialog.cpp
int DECL SetTimeDlg(CString& text, CWnd * pParentWnd, CString& title, 
      bool bMSec = false, bool b23HrLimit = false);

// SetDateTimeDlg.cpp
int DECL SetDateTimeDlg(CWnd * pParentWnd, CTime& time, SYSTEMTIME& SystemTime);

// FileNameDlg.cpp
int DECL FileNameDlg(CWnd * pParentWnd, CString& sReturn, LPCTSTR lpszDir = _T("\\"), 
      LPCTSTR lpszFilters = _T("*.*"), LPCTSTR lpszDefaultExt = _T(""), 
      LPCTSTR lpszDefaultName = NULL);

// QwertyKeypad.cpp
void DECL MakeNoise();
int DECL KeypadDlg(CWnd* pParent, CString& text, CString Caption = _T(""), CString strType = _T(""));
int DECL PasswordDlg(CWnd* pParent, CString& text);

// NumberDlg.cpp
int DECL ChangeNumberDlgReal(CWnd* pParent, realtype& rInput, LPCTSTR pszTitle = NULL);
int DECL ChangeNumberDlgInt(CWnd* pParent, int& nInput, LPCTSTR pszTitle = NULL);

// EditUnitsDlg.cpp
void DECL ChangeUnitsDlg(CWnd* pParent, CString& sUnits, BOOL& bUseNewUnits);

// MessageDlg.cpp. 
// This dialog is typically used to display a message for a limited amount of time. 
// Types: MB_OK, MB_OKCANCEL, MB_YESNO, and MB_YESNOCANCEL.
// A wait time of 0 causes dialog to wait until user presses a button. Otherwise, the 
// wait time is allowed to pass before forcing nDefaultResponse.
// Possible values for return and nDefaultResponse: IDOK, IDCANCEL, IDYES, and IDNO.
// MessageDlgCallback should not be called directly except by CGUIView.
int DECL MessageDlg(CWnd* pParent, const LPCTSTR tstrMsg, UINT nWaitSec = 0, 
      UINT nType = MB_OK, UINT nDefaultResponse = NULL);
int DECL MessageDlgCallback(CWnd* pParent, const LPCTSTR tstrMsg, UINT nWaitSec = 0, 
      UINT nType = MB_OK, UINT nDefaultResponse = NULL);

// MsgBox works just much like AfxMessageBox except it allows the caption to be set,
// and it automatically hides and restores the wait cursor.
int DECL MsgBox(LPCTSTR Prompt, UINT nType, LPCTSTR Caption);

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_MODBASE_H__199E2135_C73F_11D2_8045_00105A1E5D53__INCLUDED_)
