// i2cmgr.h : main header file for the I2CMGR DLL
//

#if !defined(AFX_I2CMGR_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_)
#define AFX_I2CMGR_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#ifndef __AFXWIN_H__
#error include 'stdafx.h' before including this file for PCH
#endif

#include "Resource.h"           // main symbols
#include "Xpert/Win32Fix.h"

#undef DECL
#ifdef ENGINE_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif


enum I2CCODE
{
   I2C_OK=0, I2C_NAK=1, I2C_TIMEOUT=2, I2C_LOSTARBITRATION=3, I2C_OVERFLOW=4,
   I2C_BUSERROR=5, I2C_RXERROR=6, I2C_SLAVETX=7, I2C_CHECKSUM=8, I2C_STOP=9, I2C_BUSBUSY=10,
   I2C_RESTART=11, I2C_BAD_CHNL=12, I2C_NUMCODES=13
};

extern "C" DECL TCHAR* I2CErrorStrings[I2C_NUMCODES];

class DECL I2CSTATS
{
public:
   I2CSTATS() { Clear(); };
   void Clear();
public:
   int RxCount;                  // Total Good Rx Counts
   int RxError;                  // Number of errors on RX attempted
   int RxFail;                   // Total number of timeouts waiting to receive a reply
   int TxCount;                  // Total Tx Good Counts
   int TxError;                  // Number of errors on TX performed
   int TxFail;                   // Number of failures on TX (after all retries have been spent)
   int Errors[I2C_NUMCODES];     // Error Counts
};


// This is the Address which sensors default to.
const I2C_DEFAULT_ADDRESS = 0xfe;

class I2CMSG;
typedef I2CMSG* PI2CMSG;

class I2CDEVICE;
typedef I2CDEVICE* PI2CDEVICE;

class I2CMGR;
typedef I2CMGR* PI2CMGR;

typedef UINT64 SERIALNO64;

typedef TCHAR SERIALNO[8];

typedef UINT8 DEVICELIST[128];

/**********************************************************************************************/
/*
I2CMSG is the primary means for storing an I2C message to be sent, or as received from a device.
Copying of messages is supported, new heap storage is allocated and free'd automatically.

Messages can be constructed 3 ways. With no initial values, with another message, or with
the components which make up a message.
*/

class DECL I2CMSG
{
public:
   I2CMSG() { ToAddr=FromAddr=Opcode=0; Data=NULL; Len=0; Time=0; Retries=0x19; Result=I2C_OK; StopBytes=0; Next=NULL; Extra=0; };
   I2CMSG(const I2CMSG& Msg) {Data=NULL; *this = Msg;};
   I2CMSG(UINT8 toAddr, UINT8 fromAddr, UINT8 opcode, PVOID data=NULL, int len=0);
   I2CMSG& operator=(const I2CMSG& Msg);
   ~I2CMSG();
public:
   DWORD   Time;        // Timestamp as to when the message was created
   PVOID   Data;        // Actual data part of a message.
   UINT16  Len;         // Length of the data
   UINT16  StopBytes;   // Stop bytes in the message
   UINT8   ToAddr;      // Who the address was or is to be sent to
   UINT8   FromAddr;    // Who the address came from
   UINT8   Retries;     // Number of Retries to attempt (priority byte of message)
   UINT8   Opcode;      // Opcode to set or received
   I2CCODE Result;      // Status code
   PI2CMSG Next;        // Special pointer for storing a message in a que or a list.
   DWORD   Extra;       // Extra data for special purposes
};


/**********************************************************************************************/
/*
QUE is a generic template class which implements a simple first in last out que and can store
any arbitrary class which supports the copy operator and has a Next pointer.

When a message is stored memory is dynamically allocated. When you Get a message from
the QUE it is removed, and if non-NULL it is the responsibility of the thread which
retrieved the message to delete it.
*/

template <class MSG>
class QUE
{
public:
   QUE();
   ~QUE();
   HANDLE hReady;
   void Store(MSG* Msg);                                // Stores a message in the Que
   void Store(const MSG& Msg) { Store(new MSG(Msg)); }; // Makes a copy of a message and stores it
   MSG* Get();                                          // Removes a message from the Que, delete it when done
   void Flush();
private:
   CRITICAL_SECTION QueLock;
   MSG* List;
};


/**********************************************************************************************/
/*
I2CDevice is the Base Class for all I2C Devices

Defines the Notify() method which is called back when a random message arrives from an
I2C device on an Address which is being Listen'ed too. The Notify() method cannot call
back in to the Manager to send a message. It should do minimal processing and return ASAP,

The base classes Notify message does nothing, see I2CDEVICEQ which Ques all incoming messages,
and provides a signal that can be waited on by a thread.
*/

class DECL I2CDEVICE
{
public:
   I2CDEVICE()                            {NextDevice = NULL; PrevDevice = NULL;};
   virtual ~I2CDEVICE()                   {};
   virtual void Notify(const I2CMSG& Msg) {};
private:
   friend I2CMGR;
   PI2CDEVICE NextDevice;
   PI2CDEVICE PrevDevice;
};


/**********************************************************************************************/
/*
I2CDEVICEQ is a simple descendant of I2CDEVICE which stores all incoming messages in a Que,
and sets an event flag (Que.hReady) which can be used in calls to WaitForSingleObject, etc.
*/

class I2CDEVICEQ : public I2CDEVICE
{
public:
   void Notify(const I2CMSG& Msg) { Que.Store(Msg); };
   protected:
   QUE<I2CMSG> Que;
};


/**********************************************************************************************/
/*
I2CREQ is a special helper class which is used to deal with responses to master sends.
when notified of an incoming message which has the same opcode it is expecting, it simply
stores the message and raises an event to signal the thread waiting for the response.
*/

class I2CREQ : public I2CDEVICE
{
public:
   I2CREQ(I2CMGR& mgr, UINT8 addr, UINT8 opcode);
   ~I2CREQ();
   void Notify(const I2CMSG& Msg);
   I2CMSG Msg;
   HANDLE hSignal;
private:
   UINT8 Opcode;
   UINT8 Addr;
   I2CMGR& Mgr;
};

/**********************************************************************************************/
/*
The I2CMGR is the primary class of the I2C Manager DLL. It processes all incoming and outgoing
messages to the I2C port. The application is responsible for Starting and Stopping the I2CMGR
thread so that it can start processing messages. Once started, the application can Send and
Request messages. Random messages are handled by Listening to an Address.

Public Methods:

Listen()

Listen to all messages from the passed in address. Address 0xfe is what sensors default to
on startup and hence can be used to listen in to the request address command sent out when
a device is plugged in. A descendent of the I2CDEVICE class is passed on to handle the callbacks.
The Notify() method of the I2CDEVICE class is called for each message which comes in.

Unlisten()

Stops listening to an address.

Send()

Send an I2C Message to an address. Retries are performed automatically.

Request()

Sends an I2C Message and waits for and returns a response. The opcode passed in specifies
which opcode to wait for. Retries are performed automatically.

StartThread()

Starts up the Manager and its monitor so it will start watching messages. This is the first thing
called before using any Send or Request commands.

StopThread()

Shutdown and turn off the Manager.

GetDeviceList()

Returns the number of devices on the bus, and fills up an array with the addresses of
every device on the bus.

GetDeviceCount()

Returns the number of devices on the bus.

SetMasterAddr()

Sets the address of the NextGen device.

GetMasterAddr()

Returns the address of the NextGen device.

SetSpeed()

Sets the I2C Bus speed. Options are: 1, 11, 55, and 90 ... all in kHz. 90 is the default.
*/

class DECL I2CMGR
{
public:
   I2CMGR();
   ~I2CMGR();
   void Listen(I2CDEVICE& Device, UINT8 Address);
   void Unlisten(I2CDEVICE& Device, UINT8 Address);
   I2CCODE Send(const I2CMSG& Msg, DWORD Timeout=100);
   I2CCODE Request(const I2CMSG& Msg, UINT8 Opcode, I2CMSG& Response, DWORD Timeout=5000);
   void NotifySend(I2CMSG& Msg);
   void StartThread();
   void StopThread();
   int GetDeviceList(DEVICELIST& DeviceList);
   int GetDeviceCount();
   void GetSerialNo(UINT8 Addr, SERIALNO& SerialNo);
   void SetModuleNo(UINT8 Addr, UINT8 ModNo);
   int GetModuleNo(UINT8 Addr);
   void SetMasterAddr(UINT8 Addr);
   UINT8 GetMasterAddr();
   void SetSpeed(int kHz);
   void ClearAfterSelftest();
   bool SerializeRequests(bool State);
   bool SetPowerSave(bool State);
   bool EnablePowerSave();
public: // Message Statistics
   I2CSTATS GetStats();
   void ClearStats();
private:
   void I2COn();
   void I2COff();
   void OpenPort();
   void ClosePort();
   void ResetChip();
   void ClearI2C();
   bool BusError(I2CCODE& Result);
   bool I2CWait(I2CCODE& Result);
   bool I2CWaitTx(I2CCODE& Result);
   void SetupI2C();
   void WaitAndCheck(I2CCODE& Result);
   void MasterSend(I2CMSG& Msg);
   void MasterRx(UCHAR SlaveAddr, char* InputData, int& InputLen, int InputLimit, I2CCODE& Result);
   bool MasterStart(UCHAR SlaveAddr, I2CCODE& Result);
   void MasterData(PVOID Data, int Len, I2CCODE& Result);
   void MasterStop(I2CCODE& Result);
   bool SlaveRecvBytes(PVOID Data, int Len, I2CCODE& Result);
   void SlaveRecv(I2CMSG& Msg);
   bool AssignAddress(I2CMSG& Msg);
   void Notify(UINT8 Addr, const I2CMSG& Msg);
   UINT8 FindSerialNo(SERIALNO64 Num);
   UINT8 AllocAddress(SERIALNO64 Num);
   void Error(I2CCODE Code, I2CCODE& Result);
private: // Manager Level Private Data
   UINT8  MasterAddr;                 // I2C Address of the NextGen Unit
   HANDLE hThread;
   HANDLE hHWThread;
   HANDLE hStopThread;
   HANDLE hHWStopThread;
   HANDLE hThreadStopped;
   HANDLE hHWThreadStopped;
   HANDLE hInterrupt;
   HANDLE hHWInterrupt;
   HANDLE hSendMessage;
   HANDLE hMessageSent;
   PI2CDEVICE DeviceMap[128];
   SERIALNO64 SerialNo[128];
   int ModuleNo[128];
   void ThreadLoop();
   void HWThreadLoop();
   friend DWORD BootStrap(PI2CMGR mgr);
   friend DWORD HWBootStrap(PI2CMGR mgr);
   I2CMSG& Message;                  // A message to be sent
   I2CMSG NoMessage;
   int DeviceCount;
   bool Serialize;
   bool PowerSave;
   bool AllowPowerSave;
private: // Message Statistics
   I2CSTATS Stats;
private: // Device Level Private Data
   UCHAR Status;                     // I2C Status
   UCHAR Data;                       // I2C Data
   UCHAR I2CClock;                   // I2C Clock Rate Register Value
   CRITICAL_SECTION DataLock;        // Critical section for accessing class data
   CRITICAL_SECTION SendLock;        // Critical section for performing a master send
   UINT16 Sum;                       // Incoming/Outgoing Checksum
};


extern "C" DECL I2CMGR Mgr;

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_I2CMGR_H__65E6C3BC_87B9_11D3_8080_00105A1E5D53__INCLUDED_)
