// Semaphore.h: interface for the TSemaphore class.
//

#if !defined(_TSEMAPHORE_H_)
#define _TSEMAPHORE_H_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "Power.h"

#undef DECL
#if defined(_WINDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif


// TSemaphore and TSingleLock are used to implement timed waits 
// (functionality that is normally provided by CSemaphore or CMutex 
// and CSingleLock, but which don't exist in our WCE build).

// Unlike event handles, which release all threads when a manual-reset 
// event is signaled, when a mutex is signaled, only one thread is 
// released. 


//////////////////////////////////////////////////////////////////////
// TSemaphore Class
//
class DECL TSemaphore
{
   // TSingleLock is typically used to lock and unlock on the stack.
   friend class TSingleLock;
   
   // Constructors and destructor.
   public:
   TSemaphore() {m_hSem = CreateMutex(NULL, FALSE, NULL);};
   TSemaphore(const TCHAR* szName) 
         {m_hSem = CreateMutex(NULL, FALSE, szName);};
   virtual ~TSemaphore() {CloseHandle(m_hSem);};

   // Methods
   public:
   bool Lock(DWORD dwWait) 
         {return (WAIT_OBJECT_0 == PowerMgr.WaitForSingleObject(m_hSem, dwWait));};
   bool Unlock() {return (FALSE != ReleaseMutex(m_hSem));};

   // Attributes
   protected:
   HANDLE m_hSem;
};


//////////////////////////////////////////////////////////////////////
// TSingleLock Class
//
class DECL TSingleLock
{
   // Constructors and destructor.
   public:
   TSingleLock(TSemaphore& sem, bool bUnlockOnDestroy = true) : m_sem(sem) 
   {
      m_bAutoUnlock = bUnlockOnDestroy;
      m_bOwn = false;
   };
   virtual ~TSingleLock() {if (m_bAutoUnlock && m_bOwn) m_sem.Unlock();};

   // Methods
   public:
   bool Lock(DWORD dwWait) {return (m_sem.Lock(dwWait) ? m_bOwn = true : false);};
   bool Unlock() {return m_sem.Unlock();};

   // Attributes
   protected:
   TSemaphore& m_sem;
   bool m_bAutoUnlock;
   bool m_bOwn;
};


//////////////////////////////////////////////////////////////////////
// TRWLock Class
//
class DECL TRWLock
{
   // This class is used to control read and write access to a resource. 
   // When a write is active, no other access is allowed. When a write 
   // is not active, any number of readers can gain access. Waits can be 
   // specified for both read and write requests. 
   
   // NOTE: Nested requests of the same type (writer making nested write 
   // request, or reader making read request) are fine. However, a reader 
   // requesting a write request will never succeed, and would deadlock 
   // itself if the requested timeout is INFINITE.

   // Constructors and destructor.
   public:
   TRWLock();
   virtual ~TRWLock();

   // Methods
   public:
   bool LockRead(DWORD dwWait = 0);
   void UnlockRead();
   bool LockWrite(DWORD dwWait = 0);
   void UnlockWrite();

   // Attributes
   protected:
   
   // Maintains count of current readers.
   int m_iCount;

   // Reader event, signaled when no readers are active.
   HANDLE m_hNoReaders;
   
   // Write lock.
   TSemaphore m_semWrite;

   // Synchronizes access to read counter.
   CRITICAL_SECTION m_cs;
};


//////////////////////////////////////////////////////////////////////
// TLock Class
//
class DECL TLock
{
   // This is a simple read/write lock class. It blocks all other access when the write
   // lock is asserted, and won't grant write access until all read locks are clear.
   // Allows n-readers and 1 writer. Nested access is allowed, which means
   // a thread which calls a function that does a LockWrite() can call a function which
   // does a LockRead() and the thread won't block. Nested LockWrite's and nested LockRead's
   // are also permitted.

   // Constructors and destructor.
public:
   TLock();
   ~TLock();

   // Methods
public:
   void LockRead();
   void UnLockRead();
   bool LockWrite(DWORD Timeout=INFINITE);
   void UnLockWrite();
   HANDLE NoReaders; // Is true when there are no readers and we can potentially write

protected:
   int Count;
   CRITICAL_SECTION Lock;
};


#endif // !defined(_TSEMAPHORE_H_)
